#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "openglheader.h"

#include "utilities.h"
#include "lights.h"
#include "trans.h"
#include "bezpatches.h"
#include "teapot.h"
#include "sproduct.h"
#include "app2f.h"
#include "app2fstruct.h"
#include "app2fproc.h"

static AppData appdata;

void LoadBPShaders ( BPRenderPrograms *brprog )
{
  static const char *filename[] =
    { "app2f0.vert.glsl", "app2f0.tesc.glsl", "app2f0.tese.glsl",
      "app2f0.geom.glsl", "app2f0.frag.glsl",
      "app2f1.vert.glsl", "app2f1.frag.glsl" };
  static const GLuint shtype[7] =
    { GL_VERTEX_SHADER, GL_TESS_CONTROL_SHADER, GL_TESS_EVALUATION_SHADER,
      GL_GEOMETRY_SHADER, GL_FRAGMENT_SHADER,
      GL_VERTEX_SHADER, GL_FRAGMENT_SHADER };
  static const GLchar *UVarNames[] =
    { "ColourSource", "LightingModel", "NormalSource", "ModifyDepth" };
  GLuint shader_id[7];
  int    i;

  for ( i = 0; i < 7; i++ )
    shader_id[i] = CompileShaderFiles ( shtype[i], 1, &filename[i] );
  brprog->program_id[0] = LinkShaderProgram ( 5, shader_id, "0" );   
  brprog->program_id[1] = LinkShaderProgram ( 2, &shader_id[5], "1" );
  GetAccessToTransBlockUniform ( brprog->program_id[0] );
  GetAccessToLightMatUniformBlocks ( brprog->program_id[0] );
  GetAccessToBezPatchStorageBlocks ( brprog->program_id[0], true, false );
  brprog->ColourSourceLoc =
      glGetUniformLocation ( brprog->program_id[0], UVarNames[0] );
  brprog->LightingModelLoc =
      glGetUniformLocation ( brprog->program_id[0], UVarNames[1] );
  brprog->NormalSourceLoc =
      glGetUniformLocation ( brprog->program_id[0], UVarNames[2] );
  brprog->ModifyDepthLoc =
      glGetUniformLocation ( brprog->program_id[0], UVarNames[3] );
  AttachUniformTransBlockToBP ( brprog->program_id[1] );
  for ( i = 0; i < 7; i++ )
    glDeleteShader ( shader_id[i] );
  ExitIfGLError ( "LoadBPShaders" );
} /*LoadBPShaders*/

GLuint LoadMiShaders ( void )
{
  static const char *filename[] =
    { "app2f2.vert.glsl", "app2f2.frag.glsl" };
  static const GLuint shtype[] = { GL_VERTEX_SHADER, GL_FRAGMENT_SHADER };
  GLuint shader_id[2], program_id;
  int    i;

  for ( i = 0; i < 2; i++ )
    shader_id[i] = CompileShaderFiles ( shtype[i], 1, &filename[i] );
  program_id = LinkShaderProgram ( 2, shader_id, "2" );
  AttachUniformTransBlockToBP ( program_id );
  for ( i = 0; i < 2; i++ )
    glDeleteShader ( shader_id[i] );
  ExitIfGLError ( "LoadMiShaders" );
  return program_id;
} /*LoadMiShaders*/ 

void SetupTeapotMatrix ( AppData *ad )
{
#define SCF (1.0/3.0)
  M4x4Translatef ( ad->teapot_mmatrix, 0.0, 0.0, -0.6 );
  M4x4MScalef ( ad->teapot_mmatrix, SCF, SCF, SCF*4.0/3.0 );
  M4x4MRotateVfv ( ad->teapot_mmatrix, ad->model_rot_axis, ad->teapot_rot_angle );
#undef SCF
} /*SetupTeapotMatrix*/

void SetupTorusMatrix ( AppData *ad )
{
  GLfloat p[4] = {3.1,0.0,2.9,1.0}, q[4];

  M4x4RotateZf ( ad->torus_mmatrix, ad->torus_rot_angle );
  M4x4MRotateXf ( ad->torus_mmatrix, 0.5*PI );
  M4x4MScalef ( ad->torus_mmatrix, 0.1, 0.1, 0.1 );
  M4x4MultMVf ( q, ad->teapot_mmatrix, p );
  M4x4TranslateMfv ( ad->torus_mmatrix, q );
} /*SetupTorusMatrix*/

static const float zenith[3] = {0.0,0.0,1.0};

void Verticalise ( float vk[3], double *angk )
{
  float   R[16], s, c, vr[3];
  double  theta, ang;

  M4x4RotateVfv ( R, vk, *angk );
  s = -V3DotProductf ( &R[0], zenith );
  c = V3DotProductf ( &R[4], zenith ); 
  theta = atan2 ( s, c );
  V3CompRotationsf ( vr, &ang, &R[8], theta, vk, *angk );
  memcpy ( vk, vr, 3*sizeof(float) );
  *angk = ang;
} /*Verticalise*/

static void SetupViewMatrix ( AppData *ad )   
{
  M4x4RotateVfv ( ad->trans.wvm, ad->camera.viewer_rvec,
                  -ad->camera.viewer_rangle );
  M4x4MultMTVf ( ad->trans.eyepos, ad->trans.wvm, ad->camera.viewer_pos0 );
  M4x4InvTranslateMfv ( ad->trans.wvm, ad->camera.viewer_pos0 );
  SetupMirrorVPMatrices ( ad->trans.eyepos, ad->trans.reyepos,
                          ad->trans.mvm, ad->trans.mpm );
} /*SetupViewMatrix*/

void _RotateViewer ( AppData *ad, double delta_xi, double delta_eta )
{
  float   vi[3], lgt, vk[3];
  double  angi, angk;

  vi[0] = (float)delta_eta*ad->camera.rl/(float)ad->camera.win_height;
  vi[1] = (float)delta_xi*ad->camera.tb/(float)ad->camera.win_width;
  vi[2] = 0.0;
  lgt = sqrt ( V3DotProductf ( vi, vi ) );
  vi[0] /= lgt;  vi[1] /= lgt;
  angi = -0.052359878;  /* -3 stopnie */
  if ( ad->vertical )
    angi *= 0.01 + 0.99*fabs ( ad->trans.wvm[1]*zenith[0] +
             ad->trans.wvm[5]*zenith[1] + ad->trans.wvm[9]*zenith[2] );
  V3CompRotationsf ( vk, &angk, ad->camera.viewer_rvec,
                     ad->camera.viewer_rangle, vi, angi );
  if ( ad->vertical )
    Verticalise ( vk, &angk );
  memcpy ( ad->camera.viewer_rvec, vk, 3*sizeof(float) );
  ad->camera.viewer_rangle = angk;
  SetupViewMatrix ( ad );
} /*_RotateViewer*/

void RotateViewer ( double delta_xi, double delta_eta )
{
  if ( delta_xi == 0.0 && delta_eta == 0.0 )
    return;  /* natychmiast uciekamy - nie chcemy dzielic przez 0 */
  _RotateViewer ( &appdata, delta_xi, delta_eta );
} /*RotateViewer*/

static GLuint LoadMyTextures ( void )
{
  GLuint tex;

  if ( (tex = CreateMyTexture ( 1024 )) ) {
    LoadMyTextureImage ( tex, 1024, 1024,   0,   0, "jaszczur.tif" );
    LoadMyTextureImage ( tex, 1024, 1024, 512,   0, "salamandra.tif" );
    LoadMyTextureImage ( tex, 1024, 1024,   0, 512, "lis.tif" );
    LoadMyTextureImage ( tex, 1024, 1024, 512, 512, "kwiatki.tif" );
    SetupMyTextureMipmaps ( tex );
  }
  return tex;
} /*LoadMyTextures*/

void InitLights ( AppData *ad )
{
  GLfloat zenith[3] = { 0.0, 0.0, 1.0 };
  GLfloat sky[4]    = { 0.3, 0.3, 0.3, 1.0 };
  GLfloat ground[4] = { 0.03, 0.03, 0.045, 1.0 };
  GLfloat amb0[3] = { 0.2, 0.2, 0.3 };
  GLfloat amb1[4] = { 0.0, 0.0, 0.0, 1.0 };
  GLfloat dir0[3] = { 0.8, 0.8, 0.8 };
  GLfloat pos0[4] = { 0.0, 1.0, 1.0, 0.0 };
  GLfloat atn0[3] = { 1.0, 0.0, 0.0 };

  SetHemisphericalLight ( &ad->light, zenith, sky, ground );
  SetHemisphericalLightOnOff ( &ad->light, false );
  SetLightAmbient ( &ad->light, 0, amb0 );
  SetLightDirect ( &ad->light, 0, dir0 );
  SetLightPosition ( &ad->light, 0, pos0 );
  SetLightAttenuation ( &ad->light, 0, atn0 );
  SetLightOnOff ( &ad->light, 0, true );
  SetLightAmbient ( &ad->light, 1, amb1 );
  SetLightDirect ( &ad->light, 1, dir0 );
  SetLightPosition ( &ad->light, 1, pos0 );
  SetLightAttenuation ( &ad->light, 1, atn0 );
  SetLightOnOff ( &ad->light, 1, false );
} /*InitLights*/

void ConstructMyTeapot ( AppData *ad )
{
  const GLfloat MyColour[4] = { 1.0, 1.0, 1.0, 1.0 };
  const GLfloat ambr[4]  = { 0.75, 0.6, 0.2, 1.0 };
  const GLfloat diffr[4] = { 0.75, 0.6, 0.2, 1.0 };
  const GLfloat specr[4] = { 0.7, 0.7, 0.6, 1.0 };
  const GLfloat shn = 60.0, wa = 5.0, we = 5.0;
  const GLfloat txc[32][4] =
    {{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.-1,0.0,0.0},
     {0.5,0.5,0.0,0.0},{0.5,1.0,0.0,0.5},{1.0,0.5,0.5,0.0},{1.0,1.0,0.5,0.5},
     {-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},
     {-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},
     {-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},
     {-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},
     {-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},
     {-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0},{-1.0,0.0,-1.0,0.0}};

  ad->myteapot = ConstructTheTeapot ( MyColour );
  SetBezierPatchTessLevel ( ad->myteapot, ad->TessLevel );
  SetBezierPatchNVS ( ad->myteapot, ad->BezNormals );
  ad->matbuf[0] = NewUniformMatBlock ();
  SetupMaterial ( ad->matbuf[0], ambr, diffr, specr, shn, wa, we );
  if ( !GenBezierPatchTextureBlock ( ad->myteapot, &txc[0][0] ) )
    ExitOnError ( "ConstructMyTeapot" );
} /*ConstructMyTeapot*/

void ConstructMyTorus ( AppData *ad )
{
  GLfloat MyColour[4] = { 0.2, 0.3, 1.0, 1.0 };
  const GLfloat ambr[4]  = { 0.0, 1.0, 1.0, 1.0 };
  const GLfloat diffr[4] = { 0.0, 0.4, 1.0, 1.0 };
  const GLfloat specr[4] = { 0.7, 0.7, 0.7, 1.0 };
  const GLfloat shn = 20.0, wa = 2.0, we = 5.0;

  ad->mytorus = EnterTorus ( 1.0, 0.5, MyColour );
  SetBezierPatchTessLevel ( ad->mytorus, ad->TessLevel );
  SetBezierPatchNVS ( ad->mytorus, ad->BezNormals );
  ad->matbuf[1] = NewUniformMatBlock ();
  SetupMaterial ( ad->matbuf[1], ambr, diffr, specr, shn, wa, we );
} /*ConstructMyTorus*/

static void _ResizeMyWorld ( AppData *ad, int width, int height )
{
  float lr, z;

  glViewport ( 0, 0, ad->camera.win_width = width,
               ad->camera.win_height = height );
  z = ad->camera.zoom;
  lr = z*0.5533*(float)width/(float)height;
  M4x4Frustumf ( ad->trans.wpm, NULL, ad->camera.left = -lr, ad->camera.right = lr,
                 ad->camera.bottom = -0.5533*z, ad->camera.top = 0.5533*z,
                 ad->camera.near = 5.0, ad->camera.far = 15.0 );
  ad->camera.rl = 2.0*lr;  ad->camera.tb = 2.0*0.5533;
  SetupMirrorVPMatrices ( ad->trans.eyepos, ad->trans.reyepos,
                          ad->trans.mvm, ad->trans.mpm );
} /*_ResizeMyWorld*/

void ResizeMyWorld ( int width, int height )
{
  _ResizeMyWorld ( &appdata, width, height );
} /*ResizeMyWorld*/

void InitCamera ( AppData *ad, int width, int height )
{
  static const float viewer_rvec[3] = {1.0,0.0,0.0};
  static const float viewer_pos0[4] = {0.0,0.0,10.0,1.0};

  memcpy ( ad->camera.viewer_rvec, viewer_rvec, 3*sizeof(float) );
  memcpy ( ad->camera.viewer_pos0, viewer_pos0, 4*sizeof(float) );
  memcpy ( ad->trans.eyepos, viewer_pos0, 4*sizeof(GLfloat) );
  ad->camera.viewer_rangle = 0.0;
  ad->camera.zoomexp = 0;  ad->camera.zoom = 1.0;
  M4x4InvTranslatefv ( ad->trans.wvm, viewer_pos0 );
  _ResizeMyWorld ( ad, width, height );
} /*InitCamera*/

void InitMyWorld ( int argc, char *argv[], int width, int height )
{
  float axis[4] = {0.0,0.0,1.0};

  memset ( &appdata, 0, sizeof(AppData) );
  LoadBPShaders ( &appdata.brprog );
  appdata.miprog = LoadMiShaders ();
  appdata.trans.trbuf = NewUniformTransBlock ();
  appdata.light.lsbuf = NewUniformLightBlock ();
  TimerInit ();
  memcpy ( appdata.model_rot_axis, axis, 4*sizeof(float) );
  appdata.teapot_rot_angle = appdata.torus_rot_angle = 0.0;
  SetupTeapotMatrix ( &appdata );
  SetupTorusMatrix ( &appdata );
  InitCamera ( &appdata, width, height );
  appdata.TessLevel = 10;
  appdata.BezNormals = GL_TRUE;
  appdata.cnet = appdata.skeleton = appdata.animate = appdata.vertical = false;
  appdata.sample_shading = appdata.draw_torus = true;
  ConstructMyTeapot ( &appdata );
  ConstructMyTorus ( &appdata );
  ConstructMirror ( &appdata.mirror );
  InitLights ( &appdata );
  appdata.mytexture = LoadMyTextures ();
  glEnable ( GL_SAMPLE_SHADING );
  glMinSampleShading ( 1.0 );
} /*InitMyWorld*/

void DrawMyTeapot ( AppData *ad )
{
  if ( ad->skeleton )
    glPolygonMode ( GL_FRONT_AND_BACK, GL_LINE );
  else
    glPolygonMode ( GL_FRONT_AND_BACK, GL_FILL );
  glUseProgram ( ad->brprog.program_id[0] );
  glUniform1i ( ad->brprog.ColourSourceLoc, ad->brprog.colour_source );
  glUniform1i ( ad->brprog.NormalSourceLoc, ad->brprog.normal_source );
  glUniform1i ( ad->brprog.ModifyDepthLoc, ad->brprog.modify_depth );
  ChooseMaterial ( ad->matbuf[0] );
  if ( ad->brprog.colour_source == 2 ) {
    BindBezPatchTextureBuffer ( ad->myteapot );
    glBindTexture ( GL_TEXTURE_2D, ad->mytexture );
    DrawBezierPatches ( ad->myteapot );
    glBindTexture ( GL_TEXTURE_2D, 0 );
  }
  else
    DrawBezierPatches ( ad->myteapot );
} /*DrawMyTeapot*/

void DrawMyTorus ( AppData *ad )
{
  if ( ad->skeleton )
    glPolygonMode ( GL_FRONT_AND_BACK, GL_LINE );
  else
    glPolygonMode ( GL_FRONT_AND_BACK, GL_FILL );
  glUseProgram ( ad->brprog.program_id[0] );
  glUniform1i ( ad->brprog.ColourSourceLoc,
                ad->brprog.colour_source > 0 ? 1 : 0 );
  glUniform1i ( ad->brprog.NormalSourceLoc, ad->brprog.normal_source );
  glUniform1i ( ad->brprog.ModifyDepthLoc, ad->brprog.modify_depth );
  ChooseMaterial ( ad->matbuf[1] );
  DrawBezierPatches ( ad->mytorus );
} /*DrawMyTorus*/

void DrawMyCNet ( BezierPatchObjf *bp, GLuint prog_id )
{
  glUseProgram ( prog_id );
  DrawBezierNets ( bp );
} /*DrawMyCNet*/

void DrawScene ( AppData *ad )
{
  LoadMMatrix ( &ad->trans, ad->teapot_mmatrix );
  DrawMyTeapot ( ad );
  if ( ad->cnet )
    DrawMyCNet ( ad->myteapot, ad->brprog.program_id[1] );
  LoadMMatrix ( &ad->trans, ad->torus_mmatrix );
  if ( ad->draw_torus ) {
    DrawMyTorus ( ad );
    if ( ad->cnet )
      DrawMyCNet ( ad->mytorus, ad->brprog.program_id[1] );
  }
} /*DrawScene*/

void DrawSceneToMirror ( AppData *ad )
{
  glBindFramebuffer ( GL_DRAW_FRAMEBUFFER, ad->mirror.mirror_fbo );
  glViewport ( 0, 0, MIRRORTXT_W, MIRRORTXT_H );
  LoadVPMatrix ( &ad->trans, true );
  glClearColor ( 0.95, 0.95, 0.95, 1.0 );
  glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  DrawScene ( ad );
  glFlush ();
} /*DrawSceneToMirror*/

void RedrawMyWorld ( void )
{
  glEnable ( GL_DEPTH_TEST );
  DrawSceneToMirror ( &appdata );
  glBindFramebuffer ( GL_DRAW_FRAMEBUFFER, 0 );
  glViewport ( 0, 0, appdata.camera.win_width, appdata.camera.win_height );
  LoadVPMatrix ( &appdata.trans, false );
  glClearColor ( 1.0, 1.0, 1.0, 1.0 );
  glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  DrawMirror ( &appdata );
  DrawScene ( &appdata );
} /*RedrawMyWorld*/

void DeleteMyWorld ( void )
{
  glUseProgram ( 0 );
  glDeleteProgram ( appdata.brprog.program_id[0] );
  glDeleteProgram ( appdata.brprog.program_id[1] );
  glDeleteProgram ( appdata.miprog );
  glDeleteBuffers ( 1, &appdata.trans.trbuf );
  glDeleteBuffers ( 1, &appdata.light.lsbuf );
  DeleteBezierPatches ( appdata.myteapot );
  DeleteBezierPatches ( appdata.mytorus );
  glDeleteBuffers ( 2, appdata.matbuf );
  glDeleteTextures ( 1, &appdata.mytexture );
  DeleteMirror ( &appdata.mirror );
  DeleteEmptyVAO ();
#ifdef DEBUG_BUFFERS_ALLOCATION
  DumpBufferIdentifiers ();
#endif
} /*DeleteMyWorld*/

char ChangeZoom ( char inczoom )
{
  int   e;
  double z;

  e = appdata.camera.zoomexp;
  if ( inczoom ) {
    if ( (z = pow ( ZOOM_FCT, (double)(++e) )) > MAX_ZOOM )
      return false;
  }
  else {
    if ( (z = pow ( ZOOM_FCT, (double)(--e) )) < MIN_ZOOM )
      return false;
  }
  appdata.camera.zoomexp = e;  appdata.camera.zoom = z;
  ResizeMyWorld ( appdata.camera.win_width, appdata.camera.win_height );
  return true;
} /*ChangeZoom*/

char ProcessCharCommand ( char charcode )
{
  switch ( charcode ) {
case '+':
    if ( appdata.TessLevel < MAX_TESS_LEVEL ) {
      SetBezierPatchTessLevel ( appdata.myteapot, ++appdata.TessLevel );
      return true;
    }
    break;
case '-':
    if ( appdata.TessLevel > MIN_TESS_LEVEL ) {
      SetBezierPatchTessLevel ( appdata.myteapot, --appdata.TessLevel );
      return true;
    }
    break;
case 'N':  case 'n':
    SetBezierPatchNVS ( appdata.myteapot,
                        appdata.BezNormals = appdata.BezNormals == 0 );
    return true;
case 'C':  case 'c':
    appdata.cnet = !appdata.cnet;
    return true;
case 'S':  case 's':
    appdata.skeleton = !appdata.skeleton;
    return true;
case 'B':  case 'b':
    appdata.brprog.lighting_model = (appdata.brprog.lighting_model+1) % 3;
    glUseProgram ( appdata.brprog.program_id[0] );
    glUniform1i ( appdata.brprog.LightingModelLoc,
                  appdata.brprog.lighting_model );
    return true;
case ' ':
    if ( (appdata.animate = !appdata.animate) )
      TimerTic ();
    return appdata.animate;
case '0':
    appdata.brprog.colour_source = 0;
    return true;
case '1':
    appdata.brprog.colour_source = 1;
    return true;
case '2':
    appdata.brprog.colour_source = 2;
    return true;
case 'D':
    appdata.brprog.normal_source = (appdata.brprog.normal_source+1) % 7;
    return true;
case 'd':
    appdata.brprog.normal_source = (appdata.brprog.normal_source+6) % 7;
    return true;
case 'M':  case 'm':
    appdata.brprog.modify_depth = !appdata.brprog.modify_depth;
    return true;
case 'A':  case 'a':
    if ( (appdata.sample_shading = !appdata.sample_shading) ) {
      glEnable ( GL_SAMPLE_SHADING );
      glMinSampleShading ( 1.0 );
    }
    else
      glDisable ( GL_SAMPLE_SHADING );
    return true;
case 'V':  case 'v':
    if ( (appdata.vertical = !appdata.vertical) ) {
      Verticalise ( appdata.camera.viewer_rvec,
                    &appdata.camera.viewer_rangle );
      SetupViewMatrix ( &appdata );
      return true;
    }
    else
      return false;
case 'T':  case 't':
    appdata.draw_torus = !appdata.draw_torus;
    return true;
case 'H':  case 'h':
    SetHemisphericalLightOnOff ( &appdata.light, !appdata.light.hl.hlon );
    SetLightOnOff ( &appdata.light, 0, !appdata.light.hl.hlon );
    SetLightOnOff ( &appdata.light, 1, appdata.light.hl.hlon );
    return true;
default:
    break;
  }
  return false;
} /*ProcessCharCommand*/

char MoveOn ( void )
{
  double dt;

  dt = TimerTocTic ();
  if ( appdata.animate ) {
    if ( (appdata.teapot_rot_angle += ANGULAR_VELOCITY1 * dt) >= PI )
      appdata.teapot_rot_angle -= 2.0*PI;
    SetupTeapotMatrix ( &appdata );
  }
  if ( (appdata.torus_rot_angle += ANGULAR_VELOCITY2 * dt) >= PI )
    appdata.torus_rot_angle -= 2.0*PI;
  SetupTorusMatrix ( &appdata );
  return true;
} /*MoveOn*/

